@@ -7,7 +7,7 @@
 """This is the Emscripten test runner. To run some tests, specify which tests
 you want, for example
 
-  test/runner asm1.test_hello_world
+  test/runner core0.test_hello_world
 
 There are many options for which tests to run and how to run them. For details,
 see
@@ -25,18 +25,22 @@ import logging
 import math
 import operator
 import os
+import platform
 import random
 import sys
 import unittest
+import time
+from functools import cmp_to_key
 
 # Setup
 
 __rootpath__ = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
-sys.path.append(__rootpath__)
+sys.path.insert(0, __rootpath__)
 
 import jsrun
 import parallel_testsuite
 import common
+from common import errlog
 from tools import shared, config, utils
 
 
@@ -55,7 +59,9 @@ passing_core_test_modes = [
   'core3',
   'cores',
   'corez',
+  'core_2gb',
   'strict',
+  'strict_js',
   'wasm2js0',
   'wasm2js1',
   'wasm2js2',
@@ -65,6 +71,11 @@ passing_core_test_modes = [
   'asan',
   'lsan',
   'ubsan',
+  'wasm64',
+  'wasm64_v8',
+  'wasm64_4gb',
+  'esm_integration',
+  'instance',
 ]
 
 # The default core test mode, used when none is specified
@@ -90,24 +101,26 @@ misc_test_modes = [
   'wasm64',
   'wasm64l',
   'bigint',
+  'browser64',
+  'browser64_4gb',
+  'browser_2gb',
 ]
 
 
 def check_js_engines():
-  working_engines = [e for e in config.JS_ENGINES if jsrun.check_engine(e)]
-  if len(working_engines) < len(config.JS_ENGINES):
-    print('Not all the JS engines in JS_ENGINES appears to work.')
-    exit(1)
+  if not all(jsrun.check_engine(e) for e in config.JS_ENGINES):
+    errlog('Not all the JS engines in JS_ENGINES appears to work.')
+    sys.exit(1)
 
   if common.EMTEST_ALL_ENGINES:
-    print('(using ALL js engines)')
+    errlog('(using ALL js engines)')
 
 
 def get_and_import_modules():
   modules = []
-  for filename in glob.glob(os.path.join(os.path.dirname(__file__), 'test*.py')):
-    module_dir, module_file = os.path.split(filename)
-    module_name, module_ext = os.path.splitext(module_file)
+  for filename in glob.glob(os.path.join(common.TEST_ROOT, 'test*.py')):
+    module_file = os.path.basename(filename)
+    module_name = os.path.splitext(module_file)[0]
     __import__(module_name)
     modules.append(sys.modules[module_name])
   return modules
@@ -125,6 +138,21 @@ def get_all_tests(modules):
   return all_tests
 
 
+def get_crossplatform_tests(modules):
+  suites = ['core0', 'other', 'sanity'] # We don't need all versions of every test
+  crossplatform_tests = []
+  # Walk over the test suites and find the test functions with the
+  # is_crossplatform_test attribute applied by @crossplatform decorator
+  for m in modules:
+    for s in suites:
+      if hasattr(m, s):
+        testclass = getattr(m, s)
+        for funcname in dir(testclass):
+          if hasattr(getattr(testclass, funcname), 'is_crossplatform_test'):
+            crossplatform_tests.append(s + '.' + funcname)
+  return crossplatform_tests
+
+
 def tests_with_expanded_wildcards(args, all_tests):
   # Process wildcards, e.g. "browser.test_pthread_*" should expand to list all pthread tests
   new_args = []
@@ -139,7 +167,7 @@ def tests_with_expanded_wildcards(args, all_tests):
     else:
       new_args += [arg]
   if not new_args and args:
-    print('No tests found to run in set: ' + str(args))
+    errlog('No tests found to run in set: ' + str(args))
     sys.exit(1)
   return new_args
 
@@ -162,7 +190,7 @@ def skip_requested_tests(args, modules):
     if arg.startswith('skip:'):
       which = arg.split('skip:')[1]
       os.environ['EMTEST_SKIP'] = os.environ['EMTEST_SKIP'] + ' ' + which
-      print('will skip "%s"' % which, file=sys.stderr)
+      errlog(f'will skip "{which}"')
       skip_test(which, modules)
       args[i] = None
   return [a for a in args if a is not None]
@@ -190,14 +218,12 @@ def get_random_test_parameters(arg):
   relevant_modes = passing_core_test_modes
   if len(arg):
     num_str = arg
-    if arg.startswith('other'):
-      base_module = 'other'
-      relevant_modes = ['other']
-      num_str = arg.replace('other', '')
-    elif arg.startswith('browser'):
-      base_module = 'browser'
-      relevant_modes = ['browser']
-      num_str = arg.replace('browser', '')
+    for mode in passing_core_test_modes + misc_test_modes:
+      if arg.startswith(mode):
+        base_module = mode
+        relevant_modes = [mode]
+        num_str = arg.replace(mode, '')
+        break
     num_tests = int(num_str)
   return num_tests, base_module, relevant_modes
 
@@ -245,11 +271,91 @@ def error_on_legacy_suite_names(args):
       utils.exit_with_error('`%s` test suite has been replaced with `%s`', a, new)
 
 
-def load_test_suites(args, modules):
+# Creates a sorter object that sorts the test run order to find the best possible
+# order to run the tests in. Generally this is slowest-first to maximize
+# parallelization, but if running with fail-fast, then the tests with recent
+# known failure frequency are run first, followed by slowest first.
+def create_test_run_sorter(failfast):
+  previous_test_run_results = common.load_previous_test_run_results()
+
+  def read_approx_fail_freq(test_name):
+    if test_name in previous_test_run_results and 'fail_frequency' in previous_test_run_results[test_name]:
+      # Quantize the float value to relatively fine-grained buckets for sorting.
+      # This bucketization is needed to merge two competing sorting goals: we may
+      # want to fail early (so tests with previous history of failures should sort first)
+      # but we also want to run the slowest tests first.
+      # We cannot sort for both goals at the same time, so have failure frequency
+      # take priority over test runtime, and quantize the failures to distinct
+      # frequencies, to be able to then sort by test runtime inside the same failure
+      # frequency bucket.
+      NUM_BUCKETS = 20
+      return round(previous_test_run_results[test_name]['fail_frequency'] * NUM_BUCKETS) / NUM_BUCKETS
+    return 0
+
+  def sort_tests_failing_and_slowest_first_comparator(x, y):
+    x = str(x)
+    y = str(y)
+
+    # Look at the number of times this test has failed, and order by failures count first
+    # Only do this in --failfast, if we are looking to fail early. (otherwise sorting by last test run duration is more productive)
+    if failfast:
+      x_fail_freq = read_approx_fail_freq(x)
+      y_fail_freq = read_approx_fail_freq(y)
+      if x_fail_freq != y_fail_freq:
+        return y_fail_freq - x_fail_freq
+
+      # Look at the number of times this test has failed overall in any other suite, and order by failures count first
+      x_fail_freq = read_approx_fail_freq(x.split(' ')[0])
+      y_fail_freq = read_approx_fail_freq(y.split(' ')[0])
+      if x_fail_freq != y_fail_freq:
+        return y_fail_freq - x_fail_freq
+
+    if x in previous_test_run_results:
+      X = previous_test_run_results[x]
+
+      # if test Y has not been run even once, run Y before X
+      if y not in previous_test_run_results:
+        return 1
+      Y = previous_test_run_results[y]
+
+      # If both X and Y have been run before, order the tests based on what the previous result was (failures first, skips very last)
+      # N.b. it is important to sandwich all skipped tests between fails and successes. This is to maximize the chances that when
+      # a failing test is detected, then the other cores will fail-fast as well. (successful tests are run slowest-first to help
+      # scheduling)
+      order_by_result = {'errored': 0, 'failed': 1, 'expected failure': 2, 'unexpected success': 3, 'skipped': 4, 'success': 5}
+      x_result = order_by_result[X['result']]
+      y_result = order_by_result[Y['result']]
+      if x_result != y_result:
+        return x_result - y_result
+
+      # Finally, order by test duration from last run
+      if X['duration'] != Y['duration']:
+        if X['result'] == 'success':
+          # If both tests were successful tests, run the slower test first to improve parallelism
+          return Y['duration'] - X['duration']
+        else:
+          # If both tests were failing tests, run the quicker test first to improve --failfast detection time
+          return X['duration'] - Y['duration']
+
+    # if test X has not been run even once, but Y has, run X before Y
+    if y in previous_test_run_results:
+      return -1
+
+    # Neither test have been run before, so run them in alphabetical order
+    return (x > y) - (x < y)
+
+  return sort_tests_failing_and_slowest_first_comparator
+
+
+def load_test_suites(args, modules, options):
+  found_start = not options.start_at
+
   loader = unittest.TestLoader()
   error_on_legacy_suite_names(args)
   unmatched_test_names = set(args)
   suites = []
+
+  total_tests = 0
   for m in modules:
     names_in_module = []
     for name in list(unmatched_test_names):
@@ -259,31 +365,45 @@ def load_test_suites(args, modules):
         unmatched_test_names.remove(name)
       except AttributeError:
         pass
-    if len(names_in_module):
+    if names_in_module:
       loaded_tests = loader.loadTestsFromNames(sorted(names_in_module), m)
       tests = flattened_tests(loaded_tests)
-      suite = suite_for_module(m, tests)
+      suite = suite_for_module(m, tests, options)
+      if options.failing_and_slow_first:
+        tests = sorted(tests, key=cmp_to_key(create_test_run_sorter(options.failfast)))
       for test in tests:
-        suite.addTest(test)
+        if not found_start:
+          # Skip over tests until we find the start
+          if test.id().endswith(options.start_at):
+            found_start = True
+          else:
+            continue
+        for _x in range(options.repeat):
+          total_tests += 1
+          suite.addTest(test)
       suites.append((m.__name__, suite))
+  if not found_start:
+    utils.exit_with_error(f'unable to find --start-at test: {options.start_at}')
+  if total_tests == 1 or parallel_testsuite.num_cores() == 1:
+    # TODO: perhaps leave it at 2 if it was 2 before?
+    common.EMTEST_SAVE_DIR = 1
   return suites, unmatched_test_names
 
 
 def flattened_tests(loaded_tests):
   tests = []
   for subsuite in loaded_tests:
-    for test in subsuite:
-      tests.append(test)
+    tests.extend(subsuite)
   return tests
 
 
-def suite_for_module(module, tests):
-  suite_supported = module.__name__ in ('test_core', 'test_other', 'test_posixtest')
+def suite_for_module(module, tests, options):
+  suite_supported = module.__name__ in ('test_core', 'test_other', 'test_posixtest', 'test_browser', 'test_codesize')
   if not common.EMTEST_SAVE_DIR and not shared.DEBUG:
     has_multiple_tests = len(tests) > 1
     has_multiple_cores = parallel_testsuite.num_cores() > 1
     if suite_supported and has_multiple_tests and has_multiple_cores:
-      return parallel_testsuite.ParallelTestSuite(len(tests))
+      return parallel_testsuite.ParallelTestSuite(len(tests), options)
   return unittest.TestSuite()
 
 
@@ -291,82 +411,122 @@ def run_tests(options, suites):
   resultMessages = []
   num_failures = 0
 
-  print('Test suites:')
-  print([s[0] for s in suites])
+  if len(suites) > 1:
+    print('Test suites:', [s[0] for s in suites])
   # Run the discovered tests
-  testRunner = unittest.TextTestRunner(verbosity=2, failfast=options.failfast)
+
+  # We currently don't support xmlrunner on macOS M1 runner since
+  # `pip` doesn't seeem to yet have pre-built binaries for M1.
+  if os.getenv('CI') and not (utils.MACOS and platform.machine() == 'arm64'):
+    os.makedirs('out', exist_ok=True)
+    # output fd must remain open until after testRunner.run() below
+    output = open('out/test-results.xml', 'wb')
+    import xmlrunner # type: ignore
+    testRunner = xmlrunner.XMLTestRunner(output=output, verbosity=2,
+                                         failfast=options.failfast)
+    print('Writing XML test output to ' + os.path.abspath(output.name))
+  else:
+    testRunner = unittest.TextTestRunner(verbosity=2, failfast=options.failfast)
+
+  total_core_time = 0
+  run_start_time = time.perf_counter()
   for mod_name, suite in suites:
-    print('Running %s: (%s tests)' % (mod_name, suite.countTestCases()))
+    errlog('Running %s: (%s tests)' % (mod_name, suite.countTestCases()))
     res = testRunner.run(suite)
     msg = ('%s: %s run, %s errors, %s failures, %s skipped' %
            (mod_name, res.testsRun, len(res.errors), len(res.failures), len(res.skipped)))
     num_failures += len(res.errors) + len(res.failures) + len(res.unexpectedSuccesses)
     resultMessages.append(msg)
+    if hasattr(res, 'core_time'):
+      total_core_time += res.core_time
+  total_run_time = time.perf_counter() - run_start_time
+  if total_core_time > 0:
+    errlog('Total core time: %.3fs. Wallclock time: %.3fs. Parallelization: %.2fx.' % (total_core_time, total_run_time, total_core_time / total_run_time))
 
   if len(resultMessages) > 1:
-    print('====================')
-    print()
-    print('TEST SUMMARY')
+    errlog('====================')
+    errlog()
+    errlog('TEST SUMMARY')
     for msg in resultMessages:
-      print('    ' + msg)
+      errlog('    ' + msg)
+
+  if options.bell:
+    sys.stdout.write('\a')
+    sys.stdout.flush()
 
   return num_failures
 
 
-def parse_args(args):
+def parse_args():
   parser = argparse.ArgumentParser(prog='runner.py', description=__doc__)
-  parser.add_argument('--save-dir', action='store_true', default=None,
+  parser.add_argument('--save-dir', action='store_true',
                       help='Save the temporary directory used during for each '
-                           'test.  Implies --cores=1.')
+                           'test.  Implies --cores=1.  Defaults to true when running a single test')
   parser.add_argument('--no-clean', action='store_true',
                       help='Do not clean the temporary directory before each test run')
-  parser.add_argument('--verbose', '-v', action='store_true', default=None)
-  parser.add_argument('--all-engines', action='store_true', default=None)
-  parser.add_argument('--detect-leaks', action='store_true', default=None)
+  parser.add_argument('--verbose', '-v', action='store_true')
+  parser.add_argument('--all-engines', action='store_true')
+  parser.add_argument('--detect-leaks', action='store_true')
   parser.add_argument('--skip-slow', action='store_true', help='Skip tests marked as slow')
-  parser.add_argument('--cores',
+  parser.add_argument('--cores', '-j',
                       help='Set the number tests to run in parallel.  Defaults '
                            'to the number of CPU cores.', default=None)
-  parser.add_argument('--rebaseline', action='store_true', default=None,
+  parser.add_argument('--rebaseline', action='store_true',
                       help='Automatically update test expectations for tests that support it.')
   parser.add_argument('--browser',
                       help='Command to launch web browser in which to run browser tests.')
+  parser.add_argument('--headless', action='store_true',
+                      help='Run browser tests in headless mode.', default=None)
+  parser.add_argument('--browser-auto-config', type=bool, default=True,
+                      help='Use the default CI browser configuration.')
   parser.add_argument('tests', nargs='*')
-  parser.add_argument('--failfast', dest='failfast', action='store_const',
-                      const=True, default=False)
-  parser.add_argument('--force64', dest='force64', action='store_const',
-                      const=True, default=None)
+  parser.add_argument('--failfast', action='store_true')
+  parser.add_argument('--failing-and-slow-first', action='store_true', help='Run failing tests first, then sorted by slowest first. Combine with --failfast for fast fail-early CI runs.')
+  parser.add_argument('--start-at', metavar='NAME', help='Skip all tests up until <NAME>')
+  parser.add_argument('--continue', dest='_continue', action='store_true',
+                      help='Resume from the last run test.'
+                           'Useful when combined with --failfast')
+  parser.add_argument('--force64', action='store_true')
+  parser.add_argument('--crossplatform-only', action='store_true')
+  parser.add_argument('--repeat', type=int, default=1,
+                      help='Repeat each test N times (default: 1).')
+  parser.add_argument('--bell', action='store_true', help='Play a sound after the test suite finishes.')
   return parser.parse_args()
 
 
 def configure():
   common.EMTEST_BROWSER = os.getenv('EMTEST_BROWSER')
+  common.EMTEST_BROWSER_AUTO_CONFIG = os.getenv('EMTEST_BROWSER_AUTO_CONFIG')
+  common.EMTEST_HEADLESS = int(os.getenv('EMTEST_HEADLESS', '0'))
   common.EMTEST_DETECT_TEMPFILE_LEAKS = int(os.getenv('EMTEST_DETECT_TEMPFILE_LEAKS', '0'))
-  common.EMTEST_SAVE_DIR = int(os.getenv('EMTEST_SAVE_DIR', '0'))
   common.EMTEST_ALL_ENGINES = int(os.getenv('EMTEST_ALL_ENGINES', '0'))
   common.EMTEST_SKIP_SLOW = int(os.getenv('EMTEST_SKIP_SLOW', '0'))
+  common.EMTEST_SKIP_FLAKY = int(os.getenv('EMTEST_SKIP_FLAKY', '0'))
+  common.EMTEST_RETRY_FLAKY = int(os.getenv('EMTEST_RETRY_FLAKY', '0'))
   common.EMTEST_LACKS_NATIVE_CLANG = int(os.getenv('EMTEST_LACKS_NATIVE_CLANG', '0'))
   common.EMTEST_REBASELINE = int(os.getenv('EMTEST_REBASELINE', '0'))
   common.EMTEST_VERBOSE = int(os.getenv('EMTEST_VERBOSE', '0')) or shared.DEBUG
-  common.EMTEST_FORCE64 = int(os.getenv('EMTEST_FORCE64', '0'))
   if common.EMTEST_VERBOSE:
     logging.root.setLevel(logging.DEBUG)
 
   assert 'PARALLEL_SUITE_EMCC_CORES' not in os.environ, 'use EMTEST_CORES rather than PARALLEL_SUITE_EMCC_CORES'
   parallel_testsuite.NUM_CORES = os.environ.get('EMTEST_CORES') or os.environ.get('EMCC_CORES')
 
+  common.configure_test_browser()
+
+
+def main():
+  options = parse_args()
+
   # Some options make sense being set in the environment, others not-so-much.
   # TODO(sbc): eventually just make these command-line only.
   if os.getenv('EMTEST_SAVE_DIR'):
-    print('Prefer --save-dir over setting $EMTEST_SAVE_DIR')
+    errlog('ERROR: use --save-dir instead of EMTEST_SAVE_DIR=1, and --no-clean instead of EMTEST_SAVE_DIR=2')
+    return 1
   if os.getenv('EMTEST_REBASELINE'):
-    print('Prefer --rebaseline over setting $EMTEST_REBASELINE')
+    errlog('Prefer --rebaseline over setting $EMTEST_REBASELINE')
   if os.getenv('EMTEST_VERBOSE'):
-    print('Prefer --verbose over setting $EMTEST_VERBOSE')
-
-
-def main(args):
-  options = parse_args(args)
+    errlog('Prefer --verbose over setting $EMTEST_VERBOSE')
 
   # We set the environments variables here and then call configure,
   # to apply them.  This means the python's multiprocessing child
@@ -384,12 +544,13 @@ def main(args):
     os.environ[name] = value
 
   set_env('EMTEST_BROWSER', options.browser)
+  set_env('EMTEST_BROWSER_AUTO_CONFIG', options.browser_auto_config)
+  set_env('EMTEST_HEADLESS', options.headless)
   set_env('EMTEST_DETECT_TEMPFILE_LEAKS', options.detect_leaks)
-  set_env('EMTEST_SAVE_DIR', options.save_dir)
+  if options.save_dir:
+    common.EMTEST_SAVE_DIR = 1
   if options.no_clean:
-    set_env('EMTEST_SAVE_DIR', 2)
-  else:
-    set_env('EMTEST_SAVE_DIR', options.save_dir)
+    common.EMTEST_SAVE_DIR = 2
   set_env('EMTEST_SKIP_SLOW', options.skip_slow)
   set_env('EMTEST_ALL_ENGINES', options.all_engines)
   set_env('EMTEST_REBASELINE', options.rebaseline)
@@ -401,6 +562,11 @@ def main(args):
 
   check_js_engines()
 
+  # Remove any old test files before starting the run
+  utils.delete_file(common.flaky_tests_log_filename)
+  utils.delete_file(common.browser_spawn_lock_filename)
+  utils.delete_file(f'{common.browser_spawn_lock_filename}_counter')
+
   def prepend_default(arg):
     if arg.startswith('test_'):
       return default_core_test_mode + '.' + arg
@@ -410,12 +576,25 @@ def main(args):
 
   modules = get_and_import_modules()
   all_tests = get_all_tests(modules)
-  tests = tests_with_expanded_wildcards(tests, all_tests)
-  tests = skip_requested_tests(tests, modules)
-  tests = args_for_random_tests(tests, modules)
-  suites, unmatched_tests = load_test_suites(tests, modules)
+  if options.crossplatform_only:
+    tests = get_crossplatform_tests(modules)
+    skip_requested_tests(options.tests, modules)
+  else:
+    tests = tests_with_expanded_wildcards(tests, all_tests)
+    tests = skip_requested_tests(tests, modules)
+    tests = args_for_random_tests(tests, modules)
+
+  if not tests:
+    errlog('ERROR: no tests to run')
+    return 1
+
+  if not options.start_at and options._continue:
+    if os.path.exists(common.LAST_TEST):
+      options.start_at = utils.read_file(common.LAST_TEST).strip()
+
+  suites, unmatched_tests = load_test_suites(tests, modules, options)
   if unmatched_tests:
-    print('ERROR: could not find the following tests: ' + ' '.join(unmatched_tests))
+    errlog('ERROR: could not find the following tests: ' + ' '.join(unmatched_tests))
     return 1
 
   num_failures = run_tests(options, suites)
@@ -429,7 +608,7 @@ configure()
 
 if __name__ == '__main__':
   try:
-    sys.exit(main(sys.argv))
+    sys.exit(main())
   except KeyboardInterrupt:
     logger.warning('KeyboardInterrupt')
     sys.exit(1)
